from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import ImpulseResponseData


def read_impulse_response(name: str) -> ImpulseResponseData:
    """Read impulse response data using torchaudio

    :param name: File name
    :return: ImpulseResponseData object
    """
    import numpy as np
    import torch
    import torchaudio

    from sonusai import SonusAIError
    from .tokenized_shell_vars import tokenized_expand

    expanded_name, _ = tokenized_expand(name)

    # Read impulse response data from audio file
    try:
        raw, sample_rate = torchaudio.load(expanded_name, backend='soundfile')
    except Exception as e:
        if name != expanded_name:
            raise SonusAIError(f'Error reading {name} (expanded: {expanded_name}): {e}')
        else:
            raise SonusAIError(f'Error reading {name}: {e}')

    raw = torch.squeeze(raw[0, :])
    offset = torch.argmax(raw)
    raw = raw[offset:]
    # Inexplicably, torch.linalg.vector_norm() causes multiprocessing contexts to hang.
    # Use np.linalg.norm() instead.
    # raw = raw / torch.linalg.vector_norm(raw)

    data = np.array(raw).astype(np.float32)
    data = data / np.linalg.norm(data)

    return ImpulseResponseData(name=name, sample_rate=sample_rate, data=data)


def read_audio(name: str) -> AudioT:
    """Read audio data from a file using torchaudio

    :param name: File name
    :return: Array of time domain audio data
    """
    import numpy as np
    import torch
    import torchaudio

    from sonusai import SonusAIError
    from .constants import SAMPLE_RATE
    from .tokenized_shell_vars import tokenized_expand

    expanded_name, _ = tokenized_expand(name)

    try:
        out, samplerate = torchaudio.load(expanded_name, backend='soundfile')
        out = torch.reshape(out[0, :], (1, out.size()[1]))
        out = torchaudio.functional.resample(out,
                                             orig_freq=samplerate,
                                             new_freq=SAMPLE_RATE,
                                             resampling_method='sinc_interp_hann')
    except Exception as e:
        if name != expanded_name:
            raise SonusAIError(f'Error reading {name} (expanded: {expanded_name}):\n{e}')
        else:
            raise SonusAIError(f'Error reading {name}:\n{e}')

    result = np.squeeze(np.array(out))
    return result



"""sonusai torchl_onnx

usage: torchl_onnx [-hv] [-b BATCH] [-t TSTEPS] [-o OUTPUT] MODEL CKPT

options:
    -h, --help
    -v, --verbose                   Be verbose
    -b BATCH, --batch BATCH         Batch size [default: 1]
    -t TSTEPS, --tsteps TSTEPS      Timesteps [default: 1]
    -o OUTPUT, --output OUTPUT      Output directory.

Convert a trained Pytorch Lightning model to ONNX.  The model is specified as an
sctl_*.py model file (sctl: sonusai custom torch lightning) and a checkpoint file
for loading weights.

Inputs:
    MODEL       SonusAI Python custom model file.
    CKPT        A Pytorch Lightning checkpoint file
    BATCH       Batch size used in onnx conversion, overrides value in model ckpt. Defaults to 1.
    TSTEPS      Timestep dimension size using in onnx conversion, overrides value in model ckpt if
                the model has a timestep dimension.  Else it is ignored.

Outputs:
    OUTPUT/     A directory containing:
                    <CKPT>.onnx        Model file with batch_size and timesteps equal to provided parameters
                    <CKPT>-b1.onnx     Model file with batch_size=1 and if the timesteps dimension exists it
                                       is set to 1 (useful for real-time inference applications)
                    torchl_onnx.log

Results are written into subdirectory <MODEL>-<TIMESTAMP> unless OUTPUT is specified.

"""
from sonusai import logger


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    batch_size = args['--batch']
    timesteps = args['--tsteps']
    model_path = args['MODEL']
    ckpt_path = args['CKPT']
    output_dir = args['--output']

    from os import makedirs
    from os.path import basename, splitext
    from sonusai.utils import import_keras_model

    # Import model definition file first to check
    model_base = basename(model_path)
    model_root = splitext(model_base)[0]
    logger.info(f'Importing model from {model_base}')
    try:
        litemodule = import_keras_model(model_path)  # note works for pytorch lightning as well as keras
    except Exception as e:
        logger.exception(f'Error: could not import model from {model_path}: {e}')
        raise SystemExit(1)

    # Load checkpoint first to get hparams if available
    from torch import load as load
    ckpt_base = basename(ckpt_path)
    ckpt_root = splitext(ckpt_base)[0]
    logger.info(f'Loading checkpoint from {ckpt_base}')
    try:
        checkpoint = load(ckpt_path, map_location=lambda storage, loc: storage)
    except Exception as e:
        logger.exception(f'Error: could not load checkpoint from {ckpt_path}: {e}')
        raise SystemExit(1)

    from os.path import join, isdir, dirname, exists
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from torch import randn
    from sonusai.utils import create_ts_name

    from sonusai.utils import create_ts_name
    from torchinfo import summary

    if batch_size is not None:
        batch_size = int(batch_size)
    if batch_size != 1:
        batch_size = 1
        logger.info(f'For now prediction only supports batch_size = 1, forcing it to 1 now')

    if timesteps is not None:
        timesteps = int(timesteps)

    if output_dir is None:
        output_dir = dirname(ckpt_path)
    else:
        if not isdir(output_dir):
            makedirs(output_dir, exist_ok=True)

    ofname = join(output_dir, ckpt_root + '.onnx')
    # First try, then add date
    if exists(ofname):
        # add hour-min-sec if necessary
        from datetime import datetime
        ts = datetime.now()
        ofname = join(output_dir, ckpt_root + '-' + ts.strftime('%Y%m%d') + '.onnx')
    ofname_root = splitext(ofname)[0]

    # Setup logging file
    create_file_handler(ofname_root + '-onnx.log')
    update_console_handler(verbose)
    initial_log_messages('torchl_onnx')
    logger.info(f'Imported model from {model_base}')
    logger.info(f'Loaded checkpoint from {ckpt_base}')

    if 'hyper_parameters' in checkpoint:
        hparams = checkpoint['hyper_parameters']
        logger.info(f'Found hyper-params on checkpoint named {checkpoint["hparams_name"]} '
                    f'with {len(hparams)} total hparams.')
        if batch_size is not None and hparams['batch_size'] != batch_size:
            if batch_size != 1:
                batch_size = 1
                logger.info(f'For now prediction only supports batch_size = 1, forcing it to 1 now')
            logger.info(f'Overriding batch_size: default = {hparams["batch_size"]}; specified = {batch_size}.')
            hparams["batch_size"] = batch_size

        if timesteps is not None:
            if hparams['timesteps'] == 0 and timesteps != 0:
                logger.warning(f'Model does not contain timesteps; ignoring override.')
                timesteps = 0

            if hparams['timesteps'] != 0 and timesteps == 0:
                logger.warning(f'Model contains timesteps; ignoring override of 0, using model default.')
                timesteps = hparams['timesteps']

            if hparams['timesteps'] != timesteps:
                logger.info(f'Overriding timesteps: default = {hparams["timesteps"]}; specified = {timesteps}.')
                hparams['timesteps'] = timesteps

        logger.info(f'Building model with hparams and batch_size={batch_size}, timesteps={timesteps}')
        try:
            model = litemodule.MyHyperModel(**hparams)  # use hparams
            # litemodule.MyHyperModel.load_from_checkpoint(ckpt_name, **hparams)
        except Exception as e:
            logger.exception(f'Error: model build (MyHyperModel) in {model_base} failed: {e}')
            raise SystemExit(1)
    else:
        logger.info(f'Warning: found checkpoint with no hyper-parameters, building model with defaults')
        try:
            tmp = litemodule.MyHyperModel()  # use default hparams
        except Exception as e:
            logger.exception(f'Error: model build (MyHyperModel) in {model_base} failed: {e}')
            raise SystemExit(1)

        if batch_size is not None:
            if tmp.batch_size != batch_size:
                logger.info(f'Overriding batch_size: default = {tmp.batch_size}; specified = {batch_size}.')
        else:
            batch_size = tmp.batch_size  # inherit

        if timesteps is not None:
            if tmp.timesteps == 0 and timesteps != 0:
                logger.warning(f'Model does not contain timesteps; ignoring override.')
                timesteps = 0

            if tmp.timesteps != 0 and timesteps == 0:
                logger.warning(f'Model contains timesteps; ignoring override.')
                timesteps = tmp.timesteps

            if tmp.timesteps != timesteps:
                logger.info(f'Overriding timesteps: default = {tmp.timesteps}; specified = {timesteps}.')
        else:
            timesteps = tmp.timesteps

        logger.info(f'Building model with default hparams and batch_size= {batch_size}, timesteps={timesteps}')
        model = litemodule.MyHyperModel(timesteps=timesteps, batch_size=batch_size)

    logger.info('')
    # logger.info(summary(model))
    # from lightning.pytorch import Trainer
    # from lightning.pytorch.callbacks import ModelSummary
    # trainer = Trainer(callbacks=[ModelSummary(max_depth=2)])
    # logger.info(trainer.summarize())
    logger.info('')
    logger.info(f'feature       {model.hparams.feature}')
    logger.info(f'num_classes   {model.num_classes}')
    logger.info(f'batch_size    {model.hparams.batch_size}')
    logger.info(f'timesteps     {model.hparams.timesteps}')
    logger.info(f'flatten       {model.flatten}')
    logger.info(f'add1ch        {model.add1ch}')
    logger.info(f'truth_mutex   {model.truth_mutex}')
    logger.info(f'input_shape   {model.input_shape}')
    logger.info('')
    logger.info(f'Loading weights from {ckpt_base}')
    # model = model.load_from_checkpoint(ckpt_path)     # weights only, has problems - needs investigation
    model.load_state_dict(checkpoint["state_dict"])
    model.eval()
    insample_shape = model.input_shape
    insample_shape.insert(0, batch_size)
    input_sample = randn(insample_shape)
    logger.info(f'Creating onnx model ...')
    for m in model.modules():
        if 'instancenorm' in m.__class__.__name__.lower():
            logger.info(f'Forcing train=false for instancenorm instance {m}, {m.__class__.__name__.lower()}')
            m.train(False)
            # m.track_running_stats=True  # has problems
    model.to_onnx(file_path=ofname, input_sample=input_sample, export_params=True)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



from .aaware_whisper import aaware_whisper
from .aixplain_whisper import aixplain_whisper
from .deepgram import deepgram
from .fastwhisper import fastwhisper
from .google import google
from .whisper import whisper



"""sonusai torchl_predict

usage: torchl_predict [-hvrw] [-i MIXID] [-a ACCEL] [-p PREC] [-d DLCPU] [-m MODEL]
                     (-k CKPT) [-b BATCH] [-t TSTEPS] INPUT ...

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -i MIXID, --mixid MIXID         Mixture ID(s) to use if input is a mixture database. [default: *].
    -a ACCEL, --accelerator ACCEL   Accelerator to use in PL trainer in non-reset mode [default: auto]
    -p PREC, --precision PREC       Precision to use in PL trainer in non-reset mode. [default: 32]
    -d DLCPU, --dataloader-cpus     Number of workers/cpus for dataloader. [default: 0]
    -m MODEL, --model MODEL         PL model .py file path.
    -k CKPT, --checkpoint CKPT      PL checkpoint file with weights.
    -b BATCH, --batch BATCH         Batch size (deprecated and forced to 1). [default: 1]
    -t TSTEPS, --tsteps TSTEPS      Timesteps. If 0, dim is not included/expected in model. [default: 0]
    -r, --reset                     Reset model between each file.
    -w, --wavdbg                    Write debug .wav files of feature input, truth, and predict. [default: False]

Run PL (Pytorch Lightning) prediction with model and checkpoint input using input data from a
SonusAI mixture database.
The PL model is imported from MODEL .py file and weights loaded from checkpoint file CKPT.

Inputs:
    ACCEL       Accelerator used for PL prediction. As of PL v2.0.8:  auto, cpu, cuda, hpu, ipu, mps, tpu
    PREC        Precision used in PL prediction. PL trainer will convert model+weights to specified prec.
                As of PL v2.0.8:
                ('16-mixed', 'bf16-mixed', '32-true', '64-true', 64, 32, 16, '64', '32', '16', 'bf16')
    MODEL       Path to a .py with MyHyperModel PL model class definition
    CKPT        A PL checkpoint file with weights.
    INPUT       The input data must be one of the following:
                * directory
                  Use SonusAI mixture database directory, generate feature and truth data if not found.
                  Run prediction on the feature. The MIXID is required (or default which is *)

                * Single WAV file or glob of WAV files
                  Using the given model, generate feature data and run prediction. A model file must be
                  provided. The MIXID is ignored.

Outputs the following to tpredict-<TIMESTAMP> directory:
    <id>.h5
        dataset:    predict
    torch_predict.log

"""
from os.path import join
from typing import Any

import h5py
import torch
from lightning.pytorch.callbacks import BasePredictionWriter

from sonusai import logger
from sonusai.mixture import Feature


class CustomWriter(BasePredictionWriter):
    def __init__(self, output_dir, write_interval):
        super().__init__(write_interval)
        self.output_dir = output_dir

    def write_on_epoch_end(self, trainer, pl_module, predictions, batch_indices):
        # this will create N (num processes) files in `output_dir` each containing
        # the predictions of it's respective rank
        # torch.save(predictions, os.path.join(self.output_dir, f"predictions_{trainer.global_rank}.pt"))

        # optionally, you can also save `batch_indices` to get the information about the data index
        # from your prediction data
        num_dev = len(batch_indices)
        logger.debug(f'Num dev: {num_dev}, prediction writer global rank: {trainer.global_rank}')
        len_pred = len(predictions)  # for debug, should be num_dev
        logger.debug(f'len predictions: {len_pred}, len batch_indices0 {len(batch_indices[0])}')
        logger.debug(f'Prediction writer batch indices: {batch_indices}')

        logger.info(f'Predictions returned: {len(predictions)}, writing to .h5 files ...')
        for ndi in range(num_dev):  # iterate over list devices (num of batch groups)
            num_batches = len(batch_indices[ndi])  # num batches in dev
            for bi in range(num_batches):  # iterate over list of batches per dev
                bsz = len(batch_indices[ndi][bi])  # batch size
                for di in range(bsz):
                    gid = batch_indices[0][bi][di]
                    # gid = (bgi+1)*bi + bi
                    # gid = bgi + bi
                    logger.debug(f'{ndi}, {bi}, {di}: global id: {gid}')
                    output_name = join(self.output_dir, trainer.predict_dataloaders.dataset.mixdb.mixtures[gid].name)
                    # output_name = join(output_dir, mixdb.mixtures[i].name)
                    pdat = predictions[bi][di, None].cpu().numpy()
                    logger.debug(f'Writing predict shape {pdat.shape} to {output_name}')
                    with h5py.File(output_name, 'a') as f:
                        if 'predict' in f:
                            del f['predict']
                        f.create_dataset('predict', data=pdat)

        # output_name = join(self.output_dir,trainer.predict_dataloaders.dataset.mixdb.mixtures[0].name)
        # logger.debug(f'Writing predict shape {pdat.shape} to {output_name}')
        # torch.save(batch_indices, os.path.join(self.output_dir, f"batch_indices_{trainer.global_rank}.pt"))


def power_compress(x):
    real = x[..., 0]
    imag = x[..., 1]
    spec = torch.complex(real, imag)
    mag = torch.abs(spec)
    phase = torch.angle(spec)
    mag = mag ** 0.3
    real_compress = mag * torch.cos(phase)
    imag_compress = mag * torch.sin(phase)
    return torch.stack([real_compress, imag_compress], 1)


def power_uncompress(real, imag):
    spec = torch.complex(real, imag)
    mag = torch.abs(spec)
    phase = torch.angle(spec)
    mag = mag ** (1. / 0.3)
    real_compress = mag * torch.cos(phase)
    imag_compress = mag * torch.sin(phase)
    return torch.stack([real_compress, imag_compress], -1)


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixids = args['--mixid']
    accel = args['--accelerator']
    prec = args['--precision']
    dlcpu = int(args['--dataloader-cpus'])
    modelpath = args['--model']
    ckpt_name = args['--checkpoint']
    batch_size = args['--batch']
    timesteps = args['--tsteps']
    reset = args['--reset']
    wavdbg = args['--wavdbg']  # write .wav if true
    input_name = args['INPUT']

    from os import makedirs
    from os.path import basename
    from os.path import isdir
    from os.path import isfile
    from os.path import join
    from os.path import splitext
    from os.path import normpath
    import h5py
    # from sonusai.utils import float_to_int16

    from torchinfo import summary
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import get_feature_from_audio
    from sonusai.utils import import_keras_model
    from sonusai.mixture import read_audio
    from sonusai.utils import create_ts_name
    from sonusai.data_generator import TorchFromMixtureDatabase

    if batch_size is not None:
        batch_size = int(batch_size)
    if batch_size != 1:
        batch_size = 1
        logger.info(f'For now prediction only supports batch_size = 1, forcing it to 1 now')

    if timesteps is not None:
        timesteps = int(timesteps)

    if len(input_name) == 1 and isdir(input_name[0]):
        in_basename = basename(normpath(input_name[0]))
    else:
        in_basename = ''

    output_dir = create_ts_name('tpredict-' + in_basename)
    makedirs(output_dir, exist_ok=True)

    # Setup logging file
    logger.info(f'Created output subdirectory {output_dir}')
    create_file_handler(join(output_dir, 'torchl_predict.log'))
    update_console_handler(verbose)
    initial_log_messages('torch_predict')
    logger.info(f'torch    {torch.__version__}')

    # Load checkpoint first to get hparams if available
    try:
        checkpoint = torch.load(ckpt_name, map_location=lambda storage, loc: storage)
    except Exception as e:
        logger.exception(f'Error: could not load checkpoint from {ckpt_name}: {e}')
        raise SystemExit(1)

    # Import model definition file
    model_base = basename(modelpath)
    model_root = splitext(model_base)[0]
    logger.info(f'Importing {modelpath}')
    litemodule = import_keras_model(modelpath)

    if 'hyper_parameters' in checkpoint:
        hparams = checkpoint['hyper_parameters']
        logger.info(f'Found checkpoint file with hyper-params named {checkpoint["hparams_name"]} '
                    f'with {len(hparams)} total hparams.')
        if batch_size is not None and hparams['batch_size'] != batch_size:
            if batch_size != 1:
                batch_size = 1
                logger.info(f'For now prediction only supports batch_size = 1, forcing it to 1 now')
            logger.info(f'Overriding batch_size: default = {hparams["batch_size"]}; specified = {batch_size}.')
            hparams["batch_size"] = batch_size

        if timesteps is not None:
            if hparams['timesteps'] == 0 and timesteps != 0:
                logger.warning(f'Model does not contain timesteps; ignoring override.')
                timesteps = 0

            if hparams['timesteps'] != 0 and timesteps == 0:
                logger.warning(f'Model contains timesteps; ignoring override, using model default.')
                timesteps = hparams['timesteps']

            if hparams['timesteps'] != timesteps:
                logger.info(f'Overriding timesteps: default = {hparams["timesteps"]}; specified = {timesteps}.')
                hparams['timesteps'] = timesteps

        logger.info(f'Building model with hparams and batch_size={batch_size}, timesteps={timesteps}')
        try:
            model = litemodule.MyHyperModel(**hparams)  # use hparams
            # litemodule.MyHyperModel.load_from_checkpoint(ckpt_name, **hparams)
        except Exception as e:
            logger.exception(f'Error: model build (MyHyperModel) in {model_base} failed: {e}')
            raise SystemExit(1)
    else:
        logger.info(f'Warning: found checkpoint with no hyper-parameters, building model with defaults')
        try:
            tmp = litemodule.MyHyperModel()  # use default hparams
        except Exception as e:
            logger.exception(f'Error: model build (MyHyperModel) in {model_base} failed: {e}')
            raise SystemExit(1)

        if batch_size is not None:
            if tmp.batch_size != batch_size:
                logger.info(f'Overriding batch_size: default = {tmp.batch_size}; specified = {batch_size}.')
        else:
            batch_size = tmp.batch_size  # inherit

        if timesteps is not None:
            if tmp.timesteps == 0 and timesteps != 0:
                logger.warning(f'Model does not contain timesteps; ignoring override.')
                timesteps = 0

            if tmp.timesteps != 0 and timesteps == 0:
                logger.warning(f'Model contains timesteps; ignoring override.')
                timesteps = tmp.timesteps

            if tmp.timesteps != timesteps:
                logger.info(f'Overriding timesteps: default = {tmp.timesteps}; specified = {timesteps}.')
        else:
            timesteps = tmp.timesteps

        logger.info(f'Building model with default hparams and batch_size= {batch_size}, timesteps={timesteps}')
        model = litemodule.MyHyperModel(timesteps=timesteps, batch_size=batch_size)

    logger.info('')
    logger.info(summary(model))
    logger.info('')
    logger.info(f'feature       {model.hparams.feature}')
    logger.info(f'num_classes   {model.num_classes}')
    logger.info(f'batch_size    {model.hparams.batch_size}')
    logger.info(f'timesteps     {model.hparams.timesteps}')
    logger.info(f'flatten       {model.flatten}')
    logger.info(f'add1ch        {model.add1ch}')
    logger.info(f'truth_mutex   {model.truth_mutex}')
    logger.info(f'input_shape   {model.input_shape}')
    logger.info('')
    logger.info(f'Loading weights from {ckpt_name}')
    # model = model.load_from_checkpoint(ckpt_name)     # weights only, needs investigation
    model.load_state_dict(checkpoint["state_dict"])
    model.eval()

    logger.info('')
    # Load mixture database and setup dataloader
    if len(input_name) == 1 and isdir(input_name[0]):  # Single path to mixdb subdir
        input_name = input_name[0]
        logger.info(f'Loading mixture database from {input_name}')
        mixdb = MixtureDatabase(input_name)
        logger.info(f'Sonusai mixture db: found {mixdb.num_mixtures} mixtures with {mixdb.num_classes} classes')

        if mixdb.feature != model.hparams.feature:
            logger.warning(f'Feature in mixture database {mixdb.feature} does not match feature in model')
            # raise SystemExit(1)

        # TBD check num_classes ??

        p_mixids = mixdb.mixids_to_list(mixids)
        sampler = None
        p_datagen = TorchFromMixtureDatabase(mixdb=mixdb,
                                             mixids=p_mixids,
                                             batch_size=model.hparams.batch_size,
                                             cut_len=0,
                                             flatten=model.flatten,
                                             add1ch=model.add1ch,
                                             random_cut=False,
                                             sampler=sampler,
                                             drop_last=False,
                                             num_workers=dlcpu)

        if wavdbg:  # setup for wav write if enabled
            # Info needed to setup inverse transform
            from pyaaware import FeatureGenerator
            from pyaaware import TorchInverseTransform
            from torchaudio import save
            # from sonusai.utils import write_wav

            half = model.num_classes // 2
            fg = FeatureGenerator(feature_mode=model.hparams.feature,
                                  num_classes=model.num_classes,
                                  truth_mutex=model.truth_mutex)
            itf = TorchInverseTransform(N=fg.itransform_N,
                                        R=fg.itransform_R,
                                        bin_start=fg.bin_start,
                                        bin_end=fg.bin_end,
                                        ttype=fg.itransform_ttype)

            if mixdb.target_files[0].truth_settings[0].function == 'target_f' or \
                    mixdb.target_files[0].truth_settings[0].function == 'target_mixture_f':
                enable_truth_wav = True
            else:
                enable_truth_wav = False

            if mixdb.target_files[0].truth_settings[0].function == 'target_mixture_f':
                enable_mix_wav = True
            else:
                enable_mix_wav = False

        if reset:
            logger.info(f'Running {mixdb.num_mixtures} mixtures individually with model reset ...')
            for idx, val in enumerate(p_datagen):
                # truth = val[1]
                feature = val[0]
                with torch.no_grad():
                    ypred = model(feature)
                output_name = join(output_dir, mixdb.mixtures[idx].name)
                pdat = ypred.detach().numpy()
                if timesteps > 0:
                    logger.debug(f'In and out tsteps: {feature.shape[1]},{pdat.shape[1]}')
                logger.debug(f'Writing predict shape {pdat.shape} to {output_name}')
                with h5py.File(output_name, 'a') as f:
                    if 'predict' in f:
                        del f['predict']
                    f.create_dataset('predict', data=pdat)

                if wavdbg:
                    owav_base = splitext(output_name)[0]
                    tmp = torch.complex(ypred[..., :half], ypred[..., half:]).permute(2, 0, 1).detach()
                    predwav, _ = itf.execute_all(tmp)
                    # predwav, _ = calculate_audio_from_transform(tmp, itf, trim=True)
                    save(owav_base + '.wav', predwav.permute([1, 0]), 16000, encoding='PCM_S', bits_per_sample=16)
                    if enable_truth_wav:
                        # Note this support truth type target_f and target_mixture_f
                        tmp = torch.complex(val[0][..., :half], val[0][..., half:2 * half]).permute(2, 0, 1).detach()
                        truthwav, _ = itf.execute_all(tmp)
                        save(owav_base + '_truth.wav', truthwav.permute([1, 0]), 16000, encoding='PCM_S',
                             bits_per_sample=16)

                    if enable_mix_wav:
                        tmp = torch.complex(val[0][..., 2 * half:3 * half], val[0][..., 3 * half:]).permute(2, 0, 1)
                        mixwav, _ = itf.execute_all(tmp.detach())
                        save(owav_base + "_mix.wav", mixwav.permute([1, 0]), 16000, encoding='PCM_S',
                             bits_per_sample=16)
                        # write_wav(owav_base + "_truth.wav", truthwav, 16000)

        else:
            logger.info(f'Running {mixdb.num_mixtures} mixtures with model builtin prediction loop ...')
            from lightning.pytorch import Trainer
            pred_writer = CustomWriter(output_dir=output_dir, write_interval="epoch")
            trainer = Trainer(default_root_dir=output_dir,
                              callbacks=[pred_writer],
                              precision=prec,
                              devices='auto',
                              accelerator=accel)  # prints avail GPU, TPU, IPU, HPU and selected device
            # trainer = Trainer(default_root_dir=output_dir,
            #                   devices='auto',
            #                   accelerator='auto')  # prints avail GPU, TPU, IPU, HPU and selected device
            # logger.info(f'Strategy: {trainer.strategy.strategy_name}')  # doesn't work for ddp strategy
            logger.info(f'Accelerator stats: {trainer.accelerator.get_device_stats(device=None)}')
            logger.info(f'World size: {trainer.world_size}')
            logger.info(f'Nodes: {trainer.num_nodes}')
            logger.info(f'Devices: {trainer.accelerator.auto_device_count()}')

            # Use builtin lightning prediction loop, returns a list
            # predictions = trainer.predict(model, p_datagen)  # standard method, but no support distributed
            with torch.no_grad():
                trainer.predict(model, p_datagen)
            # predictions = model.predict_outputs
            # pred_batch_idx = model.predict_batch_idx
            # if trainer.world_size > 1:
            #     ddp_max_mem = torch.cuda.max_memory_allocated(trainer.local_rank) / 1000
            #     logger.info(f"GPU {trainer.local_rank} max memory using DDP: {ddp_max_mem:.2f} MB")
            # if not trainer.is_global_zero:
            #     return
            # logger.debug(f'type predictions: {type(predictions)}, type batch_idx: {type(pred_batch_idx)}')
            # logger.debug(f'# predictions: {len(predictions)}, # batch_idx: {len(pred_batch_idx)}')
            # logger.debug(f'{pred_batch_idx}')
            # # # all_predictions = torch.cat(predictions)   #  predictions = torch.cat(predictions).cpu()
            # # if trainer.world_size > 1:
            # #     # print(f'Predictions returned: {len(all_predictions)}')
            # #     ddp_max_mem = torch.cuda.max_memory_allocated(trainer.local_rank) / 1000
            # #     logger.info(f"GPU {trainer.local_rank} max memory using DDP: {ddp_max_mem:.2f} MB")
            # #     gathered = [None] * torch.distributed.get_world_size()
            # #     torch.distributed.all_gather_object(gathered, predictions)
            # #     torch.distributed.all_gather_object(gathered, pred_batch_idx)
            # #     torch.distributed.barrier()
            # #     if not trainer.is_global_zero:
            # #         return
            # #     predictions = sum(gathered, [])
            # #     if trainer.global_rank == 0:
            # #         logger.info(f"All predictions gathered: {len(predictions)}")
            #
            # logger.info(f'Predictions returned: {len(predictions)}, writing to .h5 files ...')
            # #for idx, mixid in enumerate(p_mixids):
            # for i in pred_batch_idx:    # note assumes batch 0:num_mix matches 0:num_mix in mixdb.mixtures
            #     # print(f'{idx}, {mixid}')
            #     output_name = join(output_dir, mixdb.mixtures[i].name)
            #     pdat = predictions[i].cpu().numpy()
            #     logger.debug(f'Writing predict shape {pdat.shape} to {output_name}')
            #     with h5py.File(output_name, 'a') as f:
            #         if 'predict' in f:
            #             del f['predict']
            #         f.create_dataset('predict', data=pdat)
            #
            #     if wavdbg:
            #         owav_base = splitext(output_name)[0]
            #         tmp = torch.complex(predictions[idx][..., :half], predictions[idx][..., half:]).permute(2, 1, 0)
            #         predwav, _ = itf.execute_all(tmp.squeeze().detach().numpy())
            #         write_wav(owav_base + ".wav", predwav.detach().numpy(), 16000)

        logger.info(f'Saved results to {output_dir}')
        return

        # if reset:
        #     # reset mode cycles through each file one at a time
        #     for mixid in mixids:
        #         feature, _ = mixdb.mixture_ft(mixid)
        #         if feature.shape[0] > 2500:
        #             print(f'Trimming input frames from {feature.shape[0]} to {2500},')
        #             feature = feature[0:2500,::]
        #         half = feature.shape[-1] // 2
        #         noisy_spec_cmplx = torch.complex(torch.tensor(feature[..., :half]),
        #                                          torch.tensor(feature[..., half:])).to(device)
        #         del feature
        #
        #         predict = _pad_and_predict(built_model=model, feature=noisy_spec_cmplx)
        #         del noisy_spec_cmplx
        #
        #         audio_est = torch_istft_olsa_hanns(predict, mixdb.it_config.N, mixdb.it_config.R).cpu()
        #         del predict
        #         output_name = join(output_dir, splitext(mixdb.mixtures[mixid].name)[0]+'.wav')
        #         print(f'Saving prediction to {output_name}')
        #         write_wav(name=output_name, audio=float_to_int16(audio_est.detach().numpy()).transpose())
        #
        #         torch.cuda.empty_cache()
        #
        #         # TBD .h5 predict file optional output file
        #         # output_name = join(output_dir, mixdb.mixtures[mixid].name)
        #         # with h5py.File(output_name, 'a') as f:
        #         #     if 'predict' in f:
        #         #         del f['predict']
        #         #     f.create_dataset(name='predict', data=predict)
        #
        # else:
        #     # Run all data at once using a data generator
        #     feature = KerasFromH5(mixdb=mixdb,
        #                           mixids=mixids,
        #                           batch_size=hypermodel.batch_size,
        #                           timesteps=hypermodel.timesteps,
        #                           flatten=hypermodel.flatten,
        #                           add1ch=hypermodel.add1ch)
        #
        #     predict = built_model.predict(feature, batch_size=hypermodel.batch_size, verbose=1)
        #     predict, _ = reshape_outputs(predict=predict, timesteps=hypermodel.timesteps)
        #
        #     # Write data to separate files
        #     for idx, mixid in enumerate(mixids):
        #         output_name = join(output_dir, mixdb.mixtures[mixid].name)
        #         with h5py.File(output_name, 'a') as f:
        #             if 'predict' in f:
        #                 del f['predict']
        #             f.create_dataset('predict', data=predict[feature.file_indices[idx]])
        #
        # logger.info(f'Saved results to {output_dir}')
        # return

    if not all(isfile(file) and splitext(file)[1] == '.wav' for file in input_name):
        logger.exception(f'Do not know how to process input from {input_name}')
        raise SystemExit(1)

    logger.info(f'Run prediction on {len(input_name):,} WAV files')
    for file in input_name:
        # Convert WAV to feature data
        audio = read_audio(file)
        feature = get_feature_from_audio(audio=audio, feature=model.feature)

        # feature, predict = _pad_and_predict(hypermodel=hypermodel,
        #                                     built_model=built_model,
        #                                     feature=feature,
        #                                     frames_per_batch=frames_per_batch)

        # clean = torch_istft_olsa_hanns(clean_spec_cmplx, mixdb.ift_config.N, mixdb.ift_config.R)

        output_name = join(output_dir, splitext(basename(file))[0] + '.h5')
        with h5py.File(output_name, 'a') as f:
            if 'feature' in f:
                del f['feature']
            f.create_dataset(name='feature', data=feature)

            # if 'predict' in f:
            #     del f['predict']
            # f.create_dataset(name='predict', data=predict)

    logger.info(f'Saved results to {output_dir}')
    del model


def _pad_and_predict(built_model: Any, feature: Feature) -> torch.Tensor:
    """
    Run prediction on feature [frames,1,bins*2] (stacked complex numpy array, stride/tsteps=1)
    Returns predict output [batch,frames,bins] in complex torch.tensor
    """
    noisy_spec = power_compress(torch.view_as_real(torch.from_numpy(feature).permute(1, 0, 2)))
    # print(f'noisy_spec type {type(noisy_spec_cmplx)}')
    # print(f'noisy_spec dtype {noisy_spec_cmplx.dtype}')
    # print(f'noisy_spec size {noisy_spec_cmplx.shape}')
    with torch.no_grad():
        est_real, est_imag = built_model(noisy_spec)  # expects in size [batch, 2, tsteps, bins]
    est_real, est_imag = est_real.permute(0, 1, 3, 2), est_imag.permute(0, 1, 3, 2)
    est_spec_uncompress = torch.view_as_complex(power_uncompress(est_real, est_imag).squeeze(1))
    # inv tf want [ch,frames,bins] complex (synonymous with [batch,tsteps,bins]), keep as torch.tensor
    predict = est_spec_uncompress.permute(0, 2, 1)  # .detach().numpy()

    return predict


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



"""sonusai vars

usage: vars [-h]

options:
   -h, --help   Display this help.

List custom SonusAI variables.

"""


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    from os import environ
    from os import getenv

    from sonusai.mixture import DEFAULT_NOISE

    print('Custom SonusAI variables:')
    print('')
    print(f'${{default_noise}}: {DEFAULT_NOISE}')
    print('')
    print('SonusAI recognized environment variables:')
    print('')
    print(f'DEEPGRAM_API_KEY {getenv("DEEPGRAM_API_KEY")}')
    print(f'GOOGLE_SPEECH_API_KEY {getenv("GOOGLE_SPEECH_API_KEY")}')
    print('')
    items = ['DEEPGRAM_API_KEY', 'GOOGLE_SPEECH_API_KEY']
    items += [item for item in environ.keys() if item.upper().startswith("AIXP_WHISPER_")]


if __name__ == '__main__':
    main()



from typing import Any


def eq_rule_is_valid(rule: Any) -> bool:
    """Check if EQ rule is valid

    An EQ rule must be a tuple of length 3 or a list of length 3 tuples.
    """

    # Must be a list or string equal to 'none'
    if isinstance(rule, str) and rule == 'none':
        return True

    if not isinstance(rule, list):
        return False

    if len(rule) != 3:
        # If the length is not 3, then all elements must also be lists
        if not all(_check_for_none(el) for el in rule):
            return False
        rules = rule
    else:
        rules = [rule]

    for r in rules:
        # Each item must be a number or string
        if not all(isinstance(el, float | int | str) for el in r):
            return False

        if isinstance(r, str) and r == 'none':
            continue

        for el in r:
            # If a string, item must start with 'rand'
            if isinstance(el, str) and not el.startswith('rand'):
                return False

    return True


def _check_for_none(rule: Any) -> bool:
    """Check if EQ rule is 'none'
    """
    if isinstance(rule, str) and rule == 'none':
        return True
    if isinstance(rule, list) and len(rule) == 3:
        return True
    return False



from pyaaware import ForwardTransform

from sonusai.mixture import AudioF
from sonusai.mixture import AudioT
from sonusai.mixture import EnergyF


def compute_energy_f(frequency_domain: AudioF = None,
                     time_domain: AudioT = None,
                     transform: ForwardTransform = None) -> EnergyF:
    """Compute the energy in each bin

    Must provide either frequency domain or time domain input. If time domain input is provided, must also provide
    ForwardTransform object to use to convert to frequency domain.

    :param frequency_domain: Frequency domain data [frames, bins]
    :param time_domain: Time domain data [samples]
    :param transform: ForwardTransform object
    :return: Frequency domain per-bin energy data [frames, bins]
    """
    import numpy as np

    from sonusai import SonusAIError

    if frequency_domain is None:
        if time_domain is None:
            raise SonusAIError('Must provide time or frequency domain input')
        if transform is None:
            raise SonusAIError('Must provide ForwardTransform object')

        frequency_domain, _ = transform.execute_all(time_domain)

    frames, bins = frequency_domain.shape
    result = np.empty((frames, bins), dtype=np.float32)

    for f in range(frames):
        for b in range(bins):
            value = frequency_domain[f, b]
            result[f, b] = np.real(value) * np.real(value) + np.imag(value) * np.imag(value)

    return result



from sonusai.utils import ASRResult
from sonusai.utils.asr_functions.data import Data


def aixplain_whisper(data: Data) -> ASRResult:
    import tempfile
    from os import getenv
    from os.path import join

    from aixplain.factories.model_factory import ModelFactory

    from sonusai import SonusAIError
    from sonusai.utils import ASRResult
    from sonusai.utils import float_to_int16
    from sonusai.utils import write_wav

    whisper_model = data.whisper_model
    if whisper_model is None:
        envvar = 'AIXP_WHISPER_' + data.whisper_model_name.upper()
        modelkey = getenv(envvar)
        if modelkey is None:
            raise SonusAIError(f'{envvar} environment variable does not exist')

        whisper_model = ModelFactory.get(modelkey)

    with tempfile.TemporaryDirectory() as tmp:
        file = join(tmp, 'asr.wav')
        write_wav(name=file, audio=float_to_int16(data.audio))

        retry = 5
        count = 0
        while True:
            try:
                results = whisper_model.run(file)
                return ASRResult(text=results['data'], confidence=results['confidence'])
            except Exception as e:
                count += 1
                print(f'Warning: aiXplain exception: {e}')
                if count >= retry:
                    raise SonusAIError(f'Whisper exception: {e.args}')


"""
aiXplain Whisper results:
{
  'completed': True,
  'data': 'The birch canoe slid on the smooth planks.',
  'usedCredits': 3.194770833333333e-05,
  'runTime': 114.029,
  'confidence': None,
  'details': [],
  'rawData': {
    'predictions': [
      ' The birch canoe slid on the smooth planks.'
    ]
  },
  'status': 'SUCCESS'
}
"""



"""sonusai torchl_train

usage: torchl_train [-hgv] (-m MODEL) (-l VLOC) [-w WEIGHTS] [-k CKPT]
                           [-e EPOCHS] [-b BATCH] [-t TSTEPS] [-p ESP] TLOC

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -m MODEL, --model MODEL         Python .py file with MyHyperModel custom PL class definition.
    -l VLOC, --vloc VLOC            Location of SonusAI mixture database to use for validation.
    -w WEIGHTS, --weights WEIGHTS   Optional PL checkpoint file for initializing model weights.
    -k CKPT, --ckpt CKPT            Optional PL checkpoint file for full resume of training.
    -e EPOCHS, --epochs EPOCHS      Number of epochs to use in training. [default: 8].
    -b BATCH, --batch BATCH         Batch size.
    -t TSTEPS, --tsteps TSTEPS      Timesteps.
    -p ESP, --patience ESP          Early stopping patience. [default: 12]
    -g, --loss-batch-log            Enable per-batch loss log. [default: False]

Train a PL (Pytorch Lightning) model defined in MODEL .py using Sonusai mixture data in TLOC.

Inputs:
    TLOC    A SonusAI mixture database directory to use for training data.
    VLOC    A SonusAI mixture database directory to use for validation data.

Results are written into subdirectory <MODEL>-<TIMESTAMP>.
Per-batch loss history, if enabled, is written to <basename>-history-lossb.npy

"""
from sonusai import logger


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    model_name = args['--model']
    weights_name = args['--weights']
    ckpt_name = args['--ckpt']
    v_name = args['--vloc']
    epochs = int(args['--epochs'])
    batch_size = args['--batch']
    timesteps = args['--tsteps']
    esp = int(args['--patience'])
    loss_batch_log = args['--loss-batch-log']
    t_name = args['TLOC']

    import warnings
    from os import makedirs
    from os.path import basename
    from os.path import join
    from os.path import splitext

    # import keras_tuner as kt

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        # from keras import backend as kb
        # from keras.callbacks import EarlyStopping

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.data_generator import TorchFromMixtureDatabase
    # from sonusai.data_generator import KerasFromH5
    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import create_ts_name
    from sonusai.utils import import_keras_model
    from lightning.pytorch import Trainer
    from pytorch_lightning.loggers.csv_logs import CSVLogger
    from lightning.pytorch.callbacks import ModelSummary
    from lightning.pytorch.callbacks import ModelCheckpoint
    from lightning.pytorch.callbacks import EarlyStopping
    from lightning.pytorch.loggers import TensorBoardLogger

    model_base = basename(model_name)
    model_root = splitext(model_base)[0]

    if batch_size is not None:
        batch_size = int(batch_size)

    if timesteps is not None:
        timesteps = int(timesteps)

    output_dir = create_ts_name(model_root)
    makedirs(output_dir, exist_ok=True)
    base_name = join(output_dir, model_root)

    # Setup logging file
    create_file_handler(join(output_dir, 'torchl_train.log'))
    update_console_handler(verbose)
    initial_log_messages('torchl_train')
    logger.info('')

    t_mixdb = MixtureDatabase(t_name)
    logger.info(f'Training: found {t_mixdb.num_mixtures} mixtures with {t_mixdb.num_classes} classes from {t_name}')

    v_mixdb = MixtureDatabase(v_name)
    logger.info(f'Validation: found {v_mixdb.num_mixtures} mixtures with {v_mixdb.num_classes} classes from {v_name}')

    # Import model definition file
    logger.info(f'Importing {model_base}')
    litemodel = import_keras_model(model_name)  # note works for PL as well as keras

    # Check overrides
    # timesteps = check_keras_overrides(model, t_mixdb.feature, t_mixdb.num_classes, timesteps, batch_size)
    # Calculate batches per epoch, use ceiling as last batch is zero extended
    # frames_per_batch = get_frames_per_batch(batch_size, timesteps)
    # batches_per_epoch = int(np.ceil(t_mixdb.total_feature_frames('*') / frames_per_batch))

    logger.info('Building and compiling model')
    try:
        model = litemodel.MyHyperModel(feature=t_mixdb.feature,
                                       # num_classes=t_mixdb.num_classes,
                                       timesteps=timesteps,
                                       batch_size=batch_size)
    except Exception as e:
        logger.exception(f'Error: building {model_base} failed: {e}')
        raise SystemExit(1)

    logger.info('')
    # built_model.summary(print_fn=logger.info)
    # logger.info(model)
    # logger.info((summary(model)))
    # logger.info(summary(hypermodel, input_size=tuple(hypermodel.input_shape)))
    logger.info('')
    logger.info(f'feature       {model.hparams.feature}')
    logger.info(f'batch_size    {model.hparams.batch_size}')
    logger.info(f'timesteps     {model.hparams.timesteps}')
    logger.info(f'num_classes   {model.num_classes}')
    logger.info(f'flatten       {model.flatten}')
    logger.info(f'add1ch        {model.add1ch}')
    logger.info(f'input_shape   {model.input_shape}')
    logger.info(f'truth_mutex   {model.truth_mutex}')
    # logger.info(f'lossf         {hypermodel.lossf}')
    # logger.info(f'optimizer     {hypermodel.configure_optimizers()}')
    logger.info('')

    t_mixid = t_mixdb.mixids_to_list()
    v_mixid = v_mixdb.mixids_to_list()

    # Use SonusAI DataGenerator to create validation feature/truth on the fly
    sampler = None  # TBD how to stratify, also see stratified_shuffle_split_mixid(t_mixdb, vsplit=0)
    t_datagen = TorchFromMixtureDatabase(mixdb=t_mixdb,
                                         mixids=t_mixid,
                                         batch_size=model.hparams.batch_size,
                                         cut_len=model.hparams.timesteps,
                                         flatten=model.flatten,
                                         add1ch=model.add1ch,
                                         random_cut=True,
                                         sampler=sampler,
                                         drop_last=True,
                                         num_workers=4)

    v_datagen = TorchFromMixtureDatabase(mixdb=v_mixdb,
                                         mixids=v_mixid,
                                         batch_size=1,
                                         cut_len=0,
                                         flatten=model.flatten,
                                         add1ch=model.add1ch,
                                         random_cut=False,
                                         sampler=sampler,
                                         drop_last=True,
                                         num_workers=0)

    csvl = CSVLogger(output_dir, name="logs", version="")
    tbl = TensorBoardLogger(output_dir, "logs", "", log_graph=True, default_hp_metric=False)
    es_cb = EarlyStopping(monitor="val_loss", min_delta=0.00, patience=esp, verbose=False, mode="min")
    ckpt_topv = ModelCheckpoint(dirpath=output_dir + '/ckpt/', save_top_k=5, monitor="val_loss",
                                mode="min", filename=model_root + "-{epoch:03d}-{val_loss:.3g}")
    # lr_monitor = LearningRateMonitor(logging_interval="step")
    ckpt_last = ModelCheckpoint(dirpath=output_dir + '/ckpt/', save_last=True)
    # lr_monitor = LearningRateMonitor(logging_interval="step")
    callbacks = [ModelSummary(max_depth=2), ckpt_topv, es_cb, ckpt_last]  # , lr_monitor]

    profiler = None  # 'advanced'
    if profiler == 'advanced':
        from lightning.pytorch.profilers import AdvancedProfiler
        profiler = AdvancedProfiler(dirpath=output_dir, filename="perf_logs")
    else:
        profiler = None

    if weights_name is not None and ckpt_name is None:
        logger.info(f'Loading weights from {weights_name}')
        model = litemodel.MyHyperModel.load_from_checkpoint(weights_name,
                                                            feature=t_mixdb.feature,
                                                            # num_classes=t_mixdb.num_classes,
                                                            timesteps=timesteps,
                                                            batch_size=batch_size)

    if ckpt_name is not None:
        logger.info(f'Loading full checkpoint and resuming training from {ckpt_name}')
        ckpt_path = ckpt_name
    else:
        ckpt_path = None

    logger.info(f'  training mixtures    {len(t_mixid)}')
    logger.info(f'  validation mixtures  {len(v_mixid)}')
    logger.info(f'Starting training with early stopping patience = {esp} ...')
    logger.info('')

    trainer = Trainer(max_epochs=epochs,
                      default_root_dir=output_dir,
                      logger=[tbl, csvl],
                      log_every_n_steps=10,
                      profiler=profiler,
                      # precision='16-mixed',
                      # accelerator="cpu",
                      # devices=4,
                      callbacks=callbacks)
    trainer.fit(model, t_datagen, v_datagen, ckpt_path=ckpt_path)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



def linear_to_db(linear: float) -> float:
    """Convert linear value to dB value
    :param linear: Linear value
    :return: dB value
    """
    import numpy as np

    return 20 * np.log10(abs(linear))


def db_to_linear(db: float) -> float:
    """Convert dB value to linear value
    :param db: dB value
    :return: Linear value
    """
    return 10 ** (db / 20)



from sonusai.mixture.datatypes import AugmentationRule
from sonusai.mixture.datatypes import AugmentationRules
from sonusai.mixture.datatypes import AugmentedTargets
from sonusai.mixture.datatypes import TargetFile
from sonusai.mixture.datatypes import TargetFiles


def balance_targets(augmented_targets: AugmentedTargets,
                    targets: TargetFiles,
                    target_augmentations: AugmentationRules,
                    class_balancing_augmentation: AugmentationRule,
                    num_classes: int,
                    truth_mutex: bool,
                    num_ir: int,
                    mixups: list[int] = None) -> tuple[AugmentedTargets, AugmentationRules]:
    import math

    from .augmentation import get_mixups
    from .datatypes import AugmentedTarget
    from .targets import get_augmented_target_ids_by_class

    first_cba_id = len(target_augmentations)

    if mixups is None:
        mixups = get_mixups(target_augmentations)

    for mixup in mixups:
        if mixup == 1:
            continue

        augmented_target_indices_by_class = get_augmented_target_ids_by_class(
            augmented_targets=augmented_targets,
            targets=targets,
            target_augmentations=target_augmentations,
            mixup=mixup,
            num_classes=num_classes,
            truth_mutex=truth_mutex)

        largest = max([len(item) for item in augmented_target_indices_by_class])
        largest = math.ceil(largest / mixup) * mixup
        for at_indices in augmented_target_indices_by_class:
            additional_augmentations_needed = largest - len(at_indices)
            target_ids = sorted(
                list(set([augmented_targets[at_index].target_id for at_index in at_indices])))

            tfi_idx = 0
            for _ in range(additional_augmentations_needed):
                target_id = target_ids[tfi_idx]
                tfi_idx = (tfi_idx + 1) % len(target_ids)
                augmentation_index, target_augmentations = _get_unused_balancing_augmentation(
                    augmented_targets=augmented_targets,
                    targets=targets,
                    target_augmentations=target_augmentations,
                    class_balancing_augmentation=class_balancing_augmentation,
                    target_id=target_id,
                    mixup=mixup,
                    num_ir=num_ir,
                    first_cba_id=first_cba_id)
                augmented_target = AugmentedTarget(target_id=target_id,
                                                   target_augmentation_id=augmentation_index)
                augmented_targets.append(augmented_target)

    return augmented_targets, target_augmentations


def _get_unused_balancing_augmentation(augmented_targets: AugmentedTargets,
                                       targets: TargetFiles,
                                       target_augmentations: AugmentationRules,
                                       class_balancing_augmentation: AugmentationRule,
                                       target_id: int,
                                       mixup: int,
                                       num_ir: int,
                                       first_cba_id: int) -> tuple[int, AugmentationRules]:
    """Get an unused balancing augmentation for a given target file index
    """
    from dataclasses import asdict

    from .augmentation import get_augmentation_rules

    balancing_augmentations = [item for item in range(len(target_augmentations)) if item >= first_cba_id]
    used_balancing_augmentations = [at.target_augmentation_id for at in augmented_targets if
                                    at.target_id == target_id and
                                    at.target_augmentation_id in balancing_augmentations]

    augmentation_indices = [item for item in balancing_augmentations if
                            item not in used_balancing_augmentations and
                            target_augmentations[item].mixup == mixup]
    if len(augmentation_indices) > 0:
        return augmentation_indices[0], target_augmentations

    class_balancing_augmentation = get_class_balancing_augmentation(target=targets[target_id],
                                                                    default_cba=class_balancing_augmentation)
    new_augmentation = get_augmentation_rules(rules=asdict(class_balancing_augmentation), num_ir=num_ir)[0]
    new_augmentation.mixup = mixup
    target_augmentations.append(new_augmentation)
    return len(target_augmentations) - 1, target_augmentations


def get_class_balancing_augmentation(target: TargetFile, default_cba: AugmentationRule) -> AugmentationRule | None:
    """ Get the class balancing augmentation rule for the given target
    """
    if target.class_balancing_augmentation is not None:
        return target.class_balancing_augmentation
    return default_cba



from multiprocessing import current_process
from multiprocessing import get_context
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Optional

from tqdm import tqdm

CONTEXT = 'fork'


def pp_tqdm_imap(func: Callable,
                 *iterables: Iterable,
                 initializer: Optional[Callable[..., None]] = None,
                 initargs: Optional[Iterable[Any]] = None,
                 progress: Optional[tqdm] = None,
                 num_cpus: Optional[int] = None,
                 total: Optional[int] = None,
                 no_par: bool = False) -> list[Any]:
    """Performs a parallel ordered imap with tqdm progress."""
    from os import cpu_count
    from typing import Sized

    if total is None:
        total = min(len(iterable) for iterable in iterables if isinstance(iterable, Sized))

    results: list[Any] = [None] * total
    n = 0
    if no_par or current_process().daemon:
        if initializer is not None:
            if initargs is not None:
                initializer(*initargs)
            else:
                initializer()

        for result in map(func, *iterables):
            results[n] = result
            n += 1
            if progress is not None:
                progress.update()
    else:
        if num_cpus is None:
            num_cpus = cpu_count()
        elif num_cpus is float:
            num_cpus = int(round(num_cpus * cpu_count()))

        if total < num_cpus:
            num_cpus = total
        with get_context(CONTEXT).Pool(processes=num_cpus, initializer=initializer, initargs=initargs) as pool:
            for result in pool.imap(func, *iterables):  # type: ignore
                results[n] = result
                n += 1
                if progress is not None:
                    progress.update()
            pool.close()
            pool.join()

    if progress is not None:
        progress.close()
    return results


pp_imap = pp_tqdm_imap



from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import TruthFunctionConfig


class Data:
    def __init__(self, target_audio: AudioT,
                 noise_audio: AudioT,
                 mixture_audio: AudioT,
                 config: TruthFunctionConfig) -> None:
        import numpy as np
        from pyaaware import AawareForwardTransform
        from pyaaware import AawareInverseTransform
        from pyaaware import FeatureGenerator

        from sonusai import SonusAIError

        self.target_audio = target_audio
        self.noise_audio = noise_audio
        self.mixture_audio = mixture_audio
        self.config = config

        fg = FeatureGenerator(feature_mode=config.feature,
                              num_classes=config.num_classes,
                              truth_mutex=config.mutex)

        self.frame_size = fg.ftransform_R

        if len(target_audio) % self.frame_size != 0:
            raise SonusAIError(f'length of target audio {len(target_audio)} '
                               f'is not a multiple of frame size {self.frame_size}')

        self.offsets = range(0, len(target_audio), self.frame_size)
        self.zero_based_indices = [x - 1 for x in config.index]
        self.target_fft = AawareForwardTransform(N=fg.ftransform_N,
                                                 R=fg.ftransform_R,
                                                 bin_start=fg.bin_start,
                                                 bin_end=fg.bin_end,
                                                 ttype=fg.ftransform_ttype)
        self.noise_fft = AawareForwardTransform(N=fg.ftransform_N,
                                                R=fg.ftransform_R,
                                                bin_start=fg.bin_start,
                                                bin_end=fg.bin_end,
                                                ttype=fg.ftransform_ttype)
        self.mixture_fft = AawareForwardTransform(N=fg.ftransform_N,
                                                  R=fg.ftransform_R,
                                                  bin_start=fg.bin_start,
                                                  bin_end=fg.bin_end,
                                                  ttype=fg.ftransform_ttype)
        self.swin = AawareInverseTransform(N=fg.itransform_N,
                                           R=fg.itransform_R,
                                           bin_start=fg.bin_start,
                                           bin_end=fg.bin_end,
                                           ttype=fg.itransform_ttype,
                                           gain=np.float32(1)).W
        self.truth = np.zeros((len(target_audio), config.num_classes), dtype=np.float32)



def max_text_width(number_of_items: int) -> int:
    """Compute maximum text width for the indices of a sequence of items.

    :param number_of_items: Total number of items in sequence
    :return: Text width of largest item index
    """
    import numpy as np

    return int(np.ceil(np.log10(number_of_items)))



# SonusAI mixture utilities
from .audio import calculate_audio_from_transform
from .audio import calculate_transform_from_audio
from .audio import get_duration
from .audio import get_next_noise
from .audio import get_num_samples
from .audio import read_audio
from .audio import read_ir
from .audio import validate_input_file
from .augmentation import apply_augmentation
from .augmentation import apply_gain
from .augmentation import apply_impulse_response
from .augmentation import augmentation_from_rule
from .augmentation import estimate_augmented_length_from_length
from .augmentation import evaluate_random_rule
from .augmentation import get_augmentation_indices_for_mixup
from .augmentation import get_augmentation_rules
from .augmentation import get_mixups
from .augmentation import pad_audio_to_length
from .class_count import get_class_count_from_mixids
from .config import get_default_config
from .config import get_impulse_response_files
from .config import get_max_class
from .config import get_noise_files
from .config import get_spectral_masks
from .config import get_target_files
from .config import load_config
from .config import raw_load_config
from .config import update_config_from_file
from .config import update_config_from_hierarchy
from .config import update_truth_settings
from .constants import BIT_DEPTH
from .constants import CHANNEL_COUNT
from .constants import DEFAULT_CONFIG
from .constants import DEFAULT_NOISE
from .constants import DEFAULT_SPEECH
from .constants import ENCODING
from .constants import FLOAT_BYTES
from .constants import RAND_PATTERN
from .constants import REQUIRED_CONFIGS
from .constants import SAMPLE_BYTES
from .constants import SAMPLE_RATE
from .constants import VALID_AUGMENTATIONS
from .constants import VALID_CONFIGS
from .constants import VALID_NOISE_MIX_MODES
from .constants import VALID_TRUTH_SETTINGS
from .datatypes import AudioF
from .datatypes import AudioT
from .datatypes import AudiosF
from .datatypes import AudiosT
from .datatypes import Augmentation
from .datatypes import AugmentationRule
from .datatypes import AugmentationRules
from .datatypes import Augmentations
from .datatypes import AugmentedTarget
from .datatypes import AugmentedTargets
from .datatypes import ClassCount
from .datatypes import EnergyF
from .datatypes import EnergyT
from .datatypes import Feature
from .datatypes import FeatureGeneratorConfig
from .datatypes import FeatureGeneratorInfo
from .datatypes import GenFTData
from .datatypes import GenMixData
from .datatypes import GeneralizedIDs
from .datatypes import ImpulseResponseData
from .datatypes import ImpulseResponseFiles
from .datatypes import ListAudiosT
from .datatypes import Mixture
from .datatypes import MixtureDatabaseConfig
from .datatypes import Mixtures
from .datatypes import NoiseFile
from .datatypes import NoiseFiles
from .datatypes import Predict
from .datatypes import Segsnr
from .datatypes import SpectralMask
from .datatypes import SpectralMasks
from .datatypes import TargetFile
from .datatypes import TargetFiles
from .datatypes import TransformConfig
from .datatypes import Truth
from .datatypes import TruthFunctionConfig
from .datatypes import TruthSetting
from .datatypes import TruthSettings
from .datatypes import UniversalSNR
from .feature import get_feature_from_audio
from .generation import generate_mixtures
from .generation import get_all_snrs_from_config
from .generation import initialize_db
from .generation import populate_asr_manifest_table
from .generation import populate_class_label_table
from .generation import populate_class_weights_threshold_table
from .generation import populate_impulse_response_file_table
from .generation import populate_mixture_table
from .generation import populate_noise_file_table
from .generation import populate_spectral_mask_table
from .generation import populate_target_file_table
from .generation import populate_top_table
from .generation import update_mixid_width
from .generation import update_mixture
from .helpers import augmented_noise_samples
from .helpers import augmented_target_samples
from .helpers import check_audio_files_exist
from .helpers import forward_transform
from .helpers import get_ft
from .helpers import get_segsnr
from .helpers import get_truth_t
from .helpers import inverse_transform
from .helpers import mixture_metadata
from .helpers import write_mixture_data
from .helpers import write_mixture_metadata
from .log_duration_and_sizes import log_duration_and_sizes
from .mapped_snr_f import calculate_mapped_snr_f
from .mapped_snr_f import calculate_snr_f_statistics
from .mixdb import MixtureDatabase
from .mixdb import db_file
from .sox_audio import Transformer
from .spectral_mask import apply_spectral_mask
from .target_class_balancing import balance_targets
from .targets import get_augmented_target_ids_by_class
from .targets import get_augmented_target_ids_for_mixup
from .targets import get_augmented_targets
from .targets import get_target_augmentations_for_mixup
from .targets import get_truth_indices_for_target
from .tokenized_shell_vars import tokenized_expand
from .tokenized_shell_vars import tokenized_replace
from .truth import get_truth_indices_for_mixid
from .truth import truth_function
from .truth import truth_reduction



from sonusai.utils import ASRResult
from sonusai.utils.asr_functions.data import Data


def aaware_whisper(data: Data) -> ASRResult:
    import tempfile
    from math import exp
    from os import getenv
    from os.path import join

    import requests

    from sonusai import SonusAIError
    from sonusai.utils import ASRResult
    from sonusai.utils import float_to_int16
    from sonusai.utils import write_wav

    url = getenv('AAWARE_WHISPER_URL')
    if url is None:
        raise SonusAIError(f'AAWARE_WHISPER_URL environment variable does not exist')
    url += '/asr?task=transcribe&language=en&encode=true&output=json'

    with tempfile.TemporaryDirectory() as tmp:
        file = join(tmp, 'asr.wav')
        write_wav(name=file, audio=float_to_int16(data.audio))

        files = {'audio_file': (file, open(file, 'rb'), 'audio/wav')}

        try:
            response = requests.post(url, files=files)
            if not response.status_code == 200:
                if response.status_code == 422:
                    raise SonusAIError(f'Validation error: {response.json()}')
                raise SonusAIError(f'Invalid response: {response.status_code}')
            result = response.json()
            return ASRResult(text=result['text'], confidence=exp(float(result['segments'][0]['avg_logprob'])))
        except Exception as e:
            raise SonusAIError(f'Aaware Whisper exception: {e.args}')


"""
Aaware Whisper Asr Webservice results:
{
  "text": " The birch canoes slid on the smooth planks.",
  "segments": [
    {
      "id": 0,
      "seek": 0,
      "start": 0.0,
      "end": 2.32,
      "text": " The birch canoes slid on the smooth planks.",
      "tokens": [
        50364, 440, 1904, 339, 393, 78, 279, 1061, 327, 322, 264, 5508, 499,
        14592, 13, 50480
      ],
      "temperature": 0.0,
      "avg_logprob": -0.385713913861443,
      "compression_ratio": 0.86,
      "no_speech_prob": 0.006166956853121519
    }
  ],
  "language": "en"
}
"""



from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import Augmentation
from sonusai.mixture.datatypes import ImpulseResponseData


def apply_augmentation(audio: AudioT, augmentation: Augmentation, frame_length: int = 1) -> AudioT:
    """Apply augmentations to audio data using SoX

    :param audio: Audio
    :param augmentation: Augmentation
    :param frame_length: Pad resulting audio to be a multiple of this
    :return: Augmented audio
    """
    from sonusai import SonusAIError
    from .augmentation import pad_audio_to_frame
    from .constants import BIT_DEPTH
    from .constants import CHANNEL_COUNT
    from .constants import ENCODING
    from .constants import SAMPLE_RATE
    from .sox_audio import Transformer

    has_effects = False

    try:
        # Apply augmentations
        tfm = Transformer()
        tfm.set_input_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT, encoding=ENCODING)
        tfm.set_output_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT, encoding=ENCODING)

        # TODO
        #  Always normalize and remove normalize from list of available augmentations
        #  Normalize to globally set level (should this be a global config parameter,
        #  or hard-coded into the script?)
        if augmentation.normalize is not None:
            tfm.norm(db_level=augmentation.normalize)
            has_effects = True

        if augmentation.gain is not None:
            tfm.gain(gain_db=augmentation.gain, normalize=False)
            has_effects = True

        if augmentation.pitch is not None:
            tfm.pitch(n_semitones=float(augmentation.pitch) / 100)
            tfm.rate(samplerate=SAMPLE_RATE)
            has_effects = True

        if augmentation.tempo is not None:
            tfm.tempo(factor=float(augmentation.tempo), audio_type='s')
            has_effects = True

        if augmentation.eq1 is not None:
            tfm.equalizer(*augmentation.eq1)
            has_effects = True

        if augmentation.eq2 is not None:
            tfm.equalizer(*augmentation.eq2)
            has_effects = True

        if augmentation.eq3 is not None:
            tfm.equalizer(*augmentation.eq3)
            has_effects = True

        if augmentation.lpf is not None:
            tfm.lowpass(frequency=augmentation.lpf)
            has_effects = True

        if has_effects:
            audio_out = tfm.build_array(input_array=audio, sample_rate_in=SAMPLE_RATE)
        else:
            audio_out = audio

    except Exception as e:
        raise SonusAIError(f'Error applying {augmentation}: {e}')

    # make sure length is multiple of frame_length
    return pad_audio_to_frame(audio=audio_out, frame_length=frame_length)


def apply_impulse_response(audio: AudioT, ir: ImpulseResponseData) -> AudioT:
    """Apply impulse response to audio data using SoX

    :param audio: Audio
    :param ir: Impulse response data
    :return: Augmented audio
    """
    import math
    import tempfile

    import numpy as np

    from sonusai import SonusAIError
    from sonusai.utils import linear_to_db
    from .constants import SAMPLE_RATE
    from .sox_audio import Transformer

    # Early exit if no ir or if all audio is zero
    if ir is None or not audio.any():
        return audio

    # Get current maximum level in dB
    max_db = linear_to_db(max(abs(audio)))

    # Convert audio to IR sample rate
    tfm = Transformer()
    tfm.set_output_format(rate=ir.sample_rate)
    audio_out = tfm.build_array(input_array=audio, sample_rate_in=SAMPLE_RATE)

    # Pad audio to align with original and give enough room for IR tail
    pad = math.ceil(ir.length / 2)
    audio_out = np.pad(array=audio_out, pad_width=(pad, pad))

    # Write coefficients to temporary file
    temp = tempfile.NamedTemporaryFile(mode='w+t')
    for d in ir.data:
        temp.write(f'{d:f}\n')
    temp.seek(0)

    # Apply IR and convert back to global sample rate
    tfm = Transformer()
    tfm.set_output_format(rate=SAMPLE_RATE)
    tfm.fir(coefficients=temp.name)
    try:
        audio_out = tfm.build_array(input_array=audio_out, sample_rate_in=ir.sample_rate)
    except Exception as e:
        raise SonusAIError(f'Error applying IR: {e}')

    temp.close()

    # Reset level to previous max value
    tfm = Transformer()
    tfm.norm(db_level=max_db)
    audio_out = tfm.build_array(input_array=audio_out, sample_rate_in=SAMPLE_RATE)

    return audio_out[:len(audio)]



from sonusai.utils import ASRResult
from sonusai.utils.asr_functions.data import Data


def google(data: Data) -> ASRResult:
    import tempfile
    from os import getenv
    from os.path import getsize
    from os.path import join
    from random import random
    from time import sleep

    import speech_recognition as sr

    from sonusai import SonusAIError
    from sonusai.utils import float_to_int16
    from sonusai.utils import human_readable_size
    from sonusai.utils import write_wav

    key = getenv('GOOGLE_SPEECH_API_KEY')
    if key is None:
        raise SonusAIError('GOOGLE_SPEECH_API_KEY environment variable does not exist')

    r = sr.Recognizer()
    with tempfile.TemporaryDirectory() as tmp:
        file = join(tmp, 'asr.wav')
        write_wav(name=file, audio=float_to_int16(data.audio))
        size = getsize(file)
        if size > 10 * 1024 * 1024:
            print(f'Warning: file size exceeds Google single request limit: {human_readable_size(size)} > 10 MB')

        with sr.AudioFile(file) as source:
            audio = r.record(source)

        try:
            retry = 5
            count = 0
            while True:
                try:
                    sleep(count * (1 + random()))
                    results = r.recognize_google(audio, key=key, show_all=True)
                    if not isinstance(results, dict) or len(results.get('alternative', [])) == 0:
                        raise ValueError
                    break
                except ValueError:
                    print(f'Warning: speech_recognition ValueError {count}\n{results}')
                    count += 1
                    if count >= retry:
                        raise SonusAIError(f'speech_recognition exception: ValueError retry count exceeded.')

            if 'confidence' in results['alternative']:
                # return alternative with highest confidence score
                best_hypothesis = max(results['alternative'], key=lambda alternative: alternative['confidence'])
            else:
                # when there is no confidence available, we arbitrarily choose the first hypothesis.
                best_hypothesis = results['alternative'][0]
            if "transcript" not in best_hypothesis:
                raise SonusAIError('speech_recognition: UnknownValueError')
            confidence = best_hypothesis.get('confidence', 0.5)
            return ASRResult(text=best_hypothesis['transcript'], confidence=confidence)
        except sr.UnknownValueError:
            return ASRResult(text='', confidence=0)
        except sr.RequestError as e:
            raise SonusAIError(f'Could not request results from Google Speech Recognition service: {e}')


"""
Google results:
{
  "result": [
    {
      "alternative": [
        {
          "transcript": "the Birch canoe slid on the smooth planks",
          "confidence": 0.94228178
        },
        {
          "transcript": "the Burj canoe slid on the smooth planks"
        },
        {
          "transcript": "the Birch canoe slid on the smooth plank"
        },
        {
          "transcript": "the Birch canoe slit on the smooth planks"
        },
        {
          "transcript": "the Birch canoes slid on the smooth planks"
        }
      ],
      "final": true
    }
  ],
  "result_index": 0
}
"""



from sonusai.mixture.constants import SAMPLE_RATE
from sonusai.mixture.datatypes import AudioT


def write_wav(name: str, audio: AudioT, sample_rate: int = SAMPLE_RATE) -> None:
    """ Write a simple, uncompressed WAV file.

    To write multiple channels, use a 2D array of shape [samples, channels].
    The bits per sample and PCM/float are determined by the data type.

    """
    import numpy as np
    import torch
    import torchaudio

    if audio.ndim == 1:
        audio = np.reshape(audio, (1, audio.shape[0]))

    torchaudio.save(name, torch.tensor(audio), sample_rate)



"""sonusai plot

usage: plot [-hve] [-i MIXID] [-m MODEL] [-l CSV] [-o OUTPUT] INPUT

options:
    -h, --help
    -v, --verbose               Be verbose.
    -i MIXID, --mixid MIXID     Mixture to plot if input is a mixture database.
    -m MODEL, --model MODEL     Trained model ONNX file.
    -l CSV, --labels CSV        Optional CSV file of class labels (from SonusAI gentcst).
    -o OUTPUT, --output OUTPUT  Optional output HDF5 file for prediction.
    -e, --energy                Use energy plots.

Plot SonusAI audio, feature, truth, and prediction data. INPUT must be one of the following:

    * WAV
      Using the given model, generate feature data and run prediction. A model file must be
      provided. The MIXID is ignored. If --energy is specified, plot predict data as energy.

    * directory
      Using the given SonusAI mixture database directory, generate feature and truth data if not found.
      Run prediction if a model is given. The MIXID is required. (--energy is ignored.)

Prediction data will be written to OUTPUT if a model file is given and OUTPUT is specified.

There will be one plot per active truth index. In addition, the top 5 prediction classes are determined and
plotted if needed (i.e., if they were not already included in the truth plots). For plots generated using a
mixture database, then the target will also be displayed. If mixup is active, then each target involved will
be added to the corresponding truth plot.

Inputs:
    MODEL   A SonusAI trained ONNX model file. If a model file is given, prediction data will be
            generated.
    INPUT   A WAV file, or
            a directory containing a SonusAI mixture database

Outputs:
    {INPUT}-plot.pdf or {INPUT}-mix{MIXID}-plot.pdf
    plot.log
    OUTPUT (if MODEL and OUTPUT are both specified)

"""

import numpy as np
from matplotlib import pyplot as plt

from sonusai import logger
from sonusai.mixture import AudioT
from sonusai.mixture import Feature
from sonusai.mixture import Predict
from sonusai.mixture import Truth


def spec_plot(mixture: AudioT,
              feature: Feature,
              predict: Predict = None,
              target: AudioT = None,
              labels: list[str] = None,
              title: str = '') -> plt.Figure:
    from sonusai.mixture import SAMPLE_RATE

    num_plots = 4 if predict is not None else 2
    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the waveform
    x_axis = np.arange(len(mixture), dtype=np.float32) / SAMPLE_RATE
    ax[0].plot(x_axis, mixture, label='Mixture')
    ax[0].set_xlim(x_axis[0], x_axis[-1])
    ax[0].set_ylim([-1.025, 1.025])
    if target is not None:
        # Plot target time-domain waveform on top of mixture
        color = 'tab:blue'
        ax[0].plot(x_axis, target, color=color, label='Target')
        ax[0].set_ylabel('magnitude', color=color)
    ax[0].set_title('Waveform')

    # Plot the spectrogram
    ax[1].imshow(np.transpose(feature), aspect='auto', interpolation='nearest', origin='lower')
    ax[1].set_title('Feature')

    if predict is not None:
        # Plot and label the model output scores for the top-scoring classes.
        mean_predict = np.mean(predict, axis=0)
        num_classes = predict.shape[-1]
        top_n = min(10, num_classes)
        top_class_indices = np.argsort(mean_predict)[::-1][:top_n]
        ax[2].imshow(np.transpose(predict[:, top_class_indices]), aspect='auto', interpolation='nearest', cmap='gray_r')
        y_ticks = range(0, top_n)
        ax[2].set_yticks(y_ticks, [labels[top_class_indices[x]] for x in y_ticks])
        ax[2].set_ylim(-0.5 + np.array([top_n, 0]))
        ax[2].set_title('Class Scores')

        # Plot the probabilities
        ax[3].plot(predict[:, top_class_indices])
        ax[3].legend(np.array(labels)[top_class_indices], loc='best')
        ax[3].set_title('Class Probabilities')

    fig.suptitle(title)

    return fig


def spec_energy_plot(mixture: AudioT,
                     feature: Feature,
                     truth_f: Truth = None,
                     predict: Predict = None) -> plt.Figure:
    from sonusai.mixture import SAMPLE_RATE

    num_plots = 2
    if truth_f is not None:
        num_plots += 1
    if predict is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the waveform
    p = 0
    x_axis = np.arange(len(mixture), dtype=np.float32) / SAMPLE_RATE
    ax[p].plot(x_axis, mixture, label='Mixture')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    ax[p].set_ylim([-1.025, 1.025])
    ax[p].set_title('Waveform')

    # Plot the spectrogram
    p += 1
    ax[p].imshow(np.transpose(feature), aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Feature')

    if truth_f is not None:
        p += 1
        ax[p].imshow(np.transpose(truth_f), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth')

    if predict is not None:
        p += 1
        ax[p].imshow(np.transpose(predict), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Predict')

    return fig


def class_plot(mixture: AudioT,
               target: AudioT = None,
               truth_f: Truth = None,
               predict: Predict = None,
               label: str = '') -> plt.Figure:
    """ Plot mixture waveform with optional prediction and/or truth together in a single plot

    The target waveform can optionally be provided, and prediction and truth can have multiple classes.

    Inputs:
      mixture       required, numpy array [samples, 1]
      target        optional, list of numpy arrays [samples, 1]
      truth_f       optional, numpy array [frames, 1]
      predict       optional, numpy array [frames, 1]
      label         optional, label name to use when plotting

    """
    from sonusai import SonusAIError
    from sonusai.mixture import SAMPLE_RATE

    if mixture.ndim != 1:
        raise SonusAIError('Too many dimensions in mixture')

    if target is not None and target.ndim != 1:
        raise SonusAIError('Too many dimensions in target')

    # Set default to 1 frame when there is no truth or predict data
    frames = 1
    if truth_f is not None and predict is not None:
        if truth_f.ndim != 1:
            raise SonusAIError('Too many dimensions in truth_f')
        t_frames = len(truth_f)

        if predict.ndim != 1:
            raise SonusAIError('Too many dimensions in predict')
        p_frames = len(predict)

        frames = min(t_frames, p_frames)
    elif truth_f is not None:
        if truth_f.ndim != 1:
            raise SonusAIError('Too many dimensions in truth_f')
        frames = len(truth_f)
    elif predict is not None:
        if predict.ndim != 1:
            raise SonusAIError('Too many dimensions in predict')
        frames = len(predict)

    samples = (len(mixture) // frames) * frames

    # x-axis in sec
    x_axis = np.arange(samples, dtype=np.float32) / SAMPLE_RATE

    fig, ax = plt.subplots(1, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the time-domain waveforms then truth/prediction on second axis
    ax.plot(x_axis, mixture[0:samples], color='mistyrose', label='Mixture')
    color = 'red'
    ax.set_xlim(x_axis[0], x_axis[-1])
    ax.set_ylim([-1.025, 1.025])
    ax.set_ylabel(f'Amplitude', color=color)
    ax.tick_params(axis='y', labelcolor=color)

    # Plot target time-domain waveform
    if target is not None:
        ax.plot(x_axis, target[0:samples], color='blue', label='Target')

    # instantiate 2nd y-axis that shares the same x-axis
    if truth_f is not None or predict is not None:
        y_label = 'Truth/Predict'
        if truth_f is None:
            y_label = 'Predict'
        if predict is None:
            y_label = 'Truth'

        ax2 = ax.twinx()

        color = 'black'
        ax2.set_xlim(x_axis[0], x_axis[-1])
        ax2.set_ylim([-0.025, 1.025])
        ax2.set_ylabel(y_label, color=color)
        ax2.tick_params(axis='y', labelcolor=color)

        if truth_f is not None:
            ax2.plot(x_axis, expand_frames_to_samples(truth_f, samples), color='green', label='Truth')

        if predict is not None:
            ax2.plot(x_axis, expand_frames_to_samples(predict, samples), color='brown', label='Predict')

    # set only on last/bottom plot
    ax.set_xlabel('time (s)')

    fig.suptitle(label)

    return fig


def expand_frames_to_samples(x: np.ndarray, samples: int) -> np.ndarray:
    samples_per_frame = samples // len(x)
    return np.reshape(np.tile(np.expand_dims(x, 1), [1, samples_per_frame]), samples)


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    from dataclasses import asdict

    from os.path import basename
    from os.path import exists
    from os.path import isdir
    from os.path import splitext

    import h5py
    from matplotlib.backends.backend_pdf import PdfPages
    from pyaaware import FeatureGenerator
    from pyaaware import Predict

    from sonusai import SonusAIError
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import FeatureGeneratorConfig
    from sonusai.mixture import get_feature_from_audio
    from sonusai.mixture import get_truth_indices_for_mixid
    from sonusai.mixture import read_audio
    from sonusai.utils import get_label_names
    from sonusai.utils import print_mixture_details

    verbose = args['--verbose']
    model_name = args['--model']
    output_name = args['--output']
    labels_name = args['--labels']
    mixid = args['--mixid']
    energy = args['--energy']
    input_name = args['INPUT']

    if mixid is not None:
        mixid = int(mixid)

    create_file_handler('plot.log')
    update_console_handler(verbose)
    initial_log_messages('plot')

    if not exists(input_name):
        raise SonusAIError(f'{input_name} does not exist')

    logger.info('')
    logger.info(f'Input:  {input_name}')
    if model_name is not None:
        logger.info(f'Model:  {model_name}')
    if output_name is not None:
        logger.info(f'Output: {output_name}')
    logger.info('')

    ext = splitext(input_name)[1]

    model = None
    target_audio = None
    truth_f = None
    t_indices = None

    if model_name is not None:
        model = Predict(model_name)

    if ext == '.wav':
        if model_name is None:
            raise SonusAIError('Must specify MODEL when input is WAV')

        mixture_audio = read_audio(input_name)
        feature = get_feature_from_audio(audio=mixture_audio, feature=model.feature)
        fg_config = FeatureGeneratorConfig(feature_mode=model.feature,
                                           num_classes=model.output_shape[-1],
                                           truth_mutex=False)
        fg = FeatureGenerator(**asdict(fg_config))
        fg_step = fg.step
        mixdb = None
        logger.debug(f'Audio samples      {len(mixture_audio)}')
        logger.debug(f'Feature shape      {feature.shape}')

    elif isdir(input_name):
        if mixid is None:
            raise SonusAIError('Must specify mixid when input is mixture database')

        mixdb = MixtureDatabase(input_name)
        fg_step = mixdb.fg_step

        print_mixture_details(mixdb=mixdb, mixid=mixid, desc_len=24, print_fn=logger.info)

        logger.info(f'Generating data for mixture {mixid}')
        mixture_audio = mixdb.mixture_mixture(mixid)
        target_audio = mixdb.mixture_target(mixid)
        feature, truth_f = mixdb.mixture_ft(mixid)
        t_indices = [x - 1 for x in get_truth_indices_for_mixid(mixdb=mixdb, mixid=mixid)]

        target_files = [mixdb.target_file(target.file_id) for target in mixdb.mixtures[mixid].targets]
        truth_functions = list(set([sub2.function for sub1 in target_files for sub2 in sub1.truth_settings]))
        energy = 'energy_f' in truth_functions or 'snr_f' in truth_functions

        logger.debug(f'Audio samples      {len(mixture_audio)}')
        logger.debug(f'Targets:')
        mixture = mixdb.mixture(mixid)
        for target in mixture.targets:
            target_file = mixdb.target_file(target.file_id)
            name = target_file.name
            duration = target_file.duration
            augmentation = target.augmentation
            logger.debug(f'  Name             {name}')
            logger.debug(f'  Duration         {duration}')
            logger.debug(f'  Augmentation     {augmentation}')

        logger.debug(f'Feature shape      {feature.shape}')
        logger.debug(f'Truth shape        {truth_f.shape}')

    else:
        raise SonusAIError(f'Unknown file type for {input_name}')

    predict = None
    labels = None
    indices = []
    if model is not None:
        logger.debug('')
        logger.info(f'Running prediction on mixture {mixid}')
        logger.debug(f'Model feature name {model.feature}')
        logger.debug(f'Model input shape  {model.input_shape}')
        logger.debug(f'Model output shape {model.output_shape}')

        if feature.shape[0] < model.input_shape[0]:
            raise SonusAIError(f'Mixture {mixid} contains {feature.shape[0]} '
                               f'frames of data which is not enough to run prediction; '
                               f'at least {model.input_shape[0]} frames are needed for this model.\n'
                               f'Consider using a model with a smaller batch size or a mixture with more data.')

        predict = model.execute(feature)

        labels = get_label_names(num_labels=predict.shape[1], file=labels_name)

        # Report the highest-scoring classes and their scores.
        p_max = np.max(predict, axis=0)
        p_indices = np.argsort(p_max)[::-1][:5]
        p_max_len = max([len(labels[i]) for i in p_indices])

        logger.info('Top 5 active prediction classes by max:')
        for p_index in p_indices:
            logger.info(f'  {labels[p_index]:{p_max_len}s} {p_max[p_index]:.3f}')
        logger.info('')

        indices = list(p_indices)

    # Add truth indices for target (if needed)
    for t_index in t_indices:
        if t_index not in indices:
            indices.append(t_index)

    base_name = basename(splitext(input_name)[0])
    if mixdb is not None:
        title = f'{input_name} Mixture {mixid}'
        pdf_name = f'{base_name}-mix{mixid}-plot.pdf'
    else:
        title = f'{input_name}'
        pdf_name = f'{base_name}-plot.pdf'

    # Original size [frames, stride, num_bands]
    # Decimate in the stride dimension
    # Reshape to get frames*decimated_stride, num_bands
    if feature.ndim != 3:
        raise SonusAIError(f'feature does not have 3 dimensions: frames, stride, num_bands')
    spectrogram = feature[:, -fg_step:, :]
    spectrogram = np.reshape(spectrogram, (spectrogram.shape[0] * spectrogram.shape[1], spectrogram.shape[2]))

    with PdfPages(pdf_name) as pdf:
        pdf.savefig(spec_plot(mixture=mixture_audio,
                              feature=spectrogram,
                              predict=predict,
                              labels=labels,
                              title=title))
        for index in indices:
            if energy:
                t_tmp = None
                if truth_f is not None:
                    t_tmp = 10 * np.log10(truth_f + np.finfo(np.float32).eps)

                p_tmp = None
                if predict is not None:
                    p_tmp = 10 * np.log10(predict + np.finfo(np.float32).eps)

                pdf.savefig(spec_energy_plot(mixture=mixture_audio,
                                             feature=spectrogram,
                                             truth_f=t_tmp,
                                             predict=p_tmp))
            else:
                p_tmp = None
                if predict is not None:
                    p_tmp = predict[:, index]

                l_tmp = None
                if labels is not None:
                    l_tmp = labels[index]

                pdf.savefig(class_plot(mixture=mixture_audio,
                                       target=target_audio[index],
                                       truth_f=truth_f[:, index],
                                       predict=p_tmp,
                                       label=l_tmp))
        logger.info(f'Wrote {pdf_name}')

    if output_name:
        with h5py.File(output_name, 'w') as f:
            f.create_dataset(name='predict', data=predict)
            logger.info(f'Wrote {output_name}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



# SonusAI general utilities
from .asl_p56 import asl_p56
from .asr import ASRResult
from .asr import calc_asr
from .braced_glob import braced_glob
from .braced_glob import braced_iglob
from .calculate_input_shape import calculate_input_shape
from .convert_string_to_number import convert_string_to_number
from .create_ts_name import create_ts_name
from .dataclass_from_dict import dataclass_from_dict
from .db import db_to_linear
from .db import linear_to_db
from .energy_f import compute_energy_f
from .engineering_number import EngineeringNumber
from .get_frames_per_batch import get_frames_per_batch
from .get_label_names import get_label_names
from .grouper import grouper
from .human_readable_size import human_readable_size
from .keras_utils import check_keras_overrides
from .keras_utils import create_onnx_from_keras
from .keras_utils import import_and_check_keras_model
from .keras_utils import import_keras_model
from .keras_utils import keras_onnx
from .max_text_width import max_text_width
from .numeric_conversion import float_to_int16
from .numeric_conversion import int16_to_float
from .onnx_utils import SonusAIMetaData
from .onnx_utils import add_sonusai_metadata
from .onnx_utils import get_sonusai_metadata
from .onnx_utils import replace_stateful_grus
from .parallel import pp_imap
from .parallel import pp_tqdm_imap
from .print_mixture_details import print_class_count
from .print_mixture_details import print_mixture_details
from .ranges import consolidate_range
from .ranges import expand_range
from .read_mixture_data import read_mixture_data
from .read_predict_data import read_predict_data
from .reshape import get_input_shape
from .reshape import get_num_classes_from_predict
from .reshape import reshape_inputs
from .reshape import reshape_outputs
from .seconds_to_hms import seconds_to_hms
from .stacked_complex import stack_complex
from .stacked_complex import stacked_complex_imag
from .stacked_complex import stacked_complex_real
from .stacked_complex import unstack_complex
from .stratified_shuffle_split import stratified_shuffle_split_mixid
from .trim_docstring import trim_docstring
from .wave import write_wav
from .yes_or_no import yes_or_no
